/**
  * The main engine which does the whole constraint system for
  * building polygons, and also the intersection algorithm.
  *
  *
  *
  * TODO:  Make sure no 3 points are colinear, no 2 points
  *        are at the same location.
  *
  *
  *
  * @author ric Plante
  */

package ep.cg;

import ep.cg.widgets.PolyCanvas;
import java.util.Vector;

public class XEngine implements ModeConstants {

  // Constants used in the step methods.
  public final static int DONE    =  1;
  public final static int OK      =  2;
  public final static int LOOPERR = -1;
  public final static int NOXERR  = -2;

  // Made public so that calipers can be drawn
  public   double ang1P_;
  public   double ang2P_;
  public Point curPt1_ = null;
  public Point curPt2_ = null;

  // Internal state
  private Polygon polygon1_;
  private Polygon polygon2_;
  private Polygon polygoni_;
  private Polygon editPoly_;
  private Point   mousePt_;
  private Point   savePt_;
  private Segment curSeg1_;
  private Segment curSeg2_;
  private Vector  pocketLids_ = null;
  private int currentMode_ = MODE_ADD1;

  // Stuff for step by step CH calculations.
  private int   curPt1idx_= 0, init1idx_ = 0;
  private int   curPt2idx_= 0, init2idx_ = 0;
  private  double ang1_;
  private  double ang2_;
  private boolean gotAroundOnce1_ = false;
  private boolean gotAroundOnce2_ = false;
  private int whichSideof1is2_ = 0;
  private int[] poly1LidIdx_ = null;
  private int[] poly2LidIdx_ = null;
  private int[] poly1XIdx_   = null;
  private int[] poly2XIdx_   = null;

  //Stuff for Xsection
  private int currentPocketNb_ = 0;
  private boolean p1IsCwiseOn1stLid_ = false;
  private boolean p1Iscwise_ = false;
  private boolean triangReady_ = false;
  private boolean triangDone_  = false;
  private boolean xSectionBuilt_ = false;
  private Segment currentLid_  = null;
  private Point   lastXsection_= null;




  public XEngine() {
    reset();
  }

  public final void reset() {
    triangDone_  = false;
    xSectionBuilt_ = false;
    polygon1_ = new Polygon();
    polygon1_.setIdent(1);
    polygon2_ = new Polygon();
    polygon2_.setIdent(2);
    polygoni_ = new Polygon();
    polygoni_.setIdent(6);
    mousePt_ = null;
    savePt_  = null;
    editPoly_ = null;
    curSeg1_  = null;
    curSeg2_  = null;
  }

  public final void setMode(int mode) {
    if ( triangDone_ ) {
      rebuildPolygon(polygon1_);
      rebuildPolygon(polygon2_);
    }
    clearCvxHull();
    clearIPolygon();
    mouseCancel();
    triangDone_ = false;
    xSectionBuilt_ = false;
    if (mode <= MODE_DEL && mode > 0)
      currentMode_ = mode;
  }

  public void mouseDown(double x, double y) {

    int n, i;
    
    if ( triangDone_ ) {
      rebuildPolygon(polygon1_);
      rebuildPolygon(polygon2_);
    }
    clearCvxHull();
    clearIPolygon();
    triangDone_ = false;
    xSectionBuilt_ = false;
    
    switch (currentMode_) {
    case MODE_ADD1:
      if (polygon1_.checkNewConvex(x,y)) {
	int psize = polygon1_.size();
	mousePt_ = new Point();
	mousePt_.setIdent(1);
	mousePt_.x( x );
	mousePt_.y( y );
	if ( psize >= 1) {
	  Point last = polygon1_.getPoint(psize-1);
	  curSeg1_ = new Segment(last, mousePt_, 1);
	  last.addSeg(curSeg1_);
	  if ( psize >= 2) {
	    Point first = polygon1_.getPoint(0);
	    last.remSeg(first);
	    curSeg2_ = new Segment(mousePt_, first, 1);
	    mousePt_.addSeg(curSeg2_);
	  }
	  else
	    curSeg2_ = null;
	}
	else
	  curSeg1_ = null;
      }
      break;
    case MODE_ADD2:
      if (polygon2_.checkNewConvex(x,y)) {
	int psize = polygon2_.size();
	mousePt_ = new Point();
	mousePt_.setIdent(2);
	mousePt_.x( x );
	mousePt_.y( y );
	if ( psize >= 1) {
	  Point last = polygon2_.getPoint(psize-1);
	  curSeg1_ = new Segment(last, mousePt_, 2);
	  last.addSeg(curSeg1_);
	  if ( psize >= 2) {
	    Point first = polygon2_.getPoint(0);
	    last.remSeg(first, 2);
	    curSeg2_ = new Segment(mousePt_, first, 2);
	    mousePt_.addSeg(curSeg2_);
	  }
	  else
	    curSeg2_ = null;
	}
	else
	  curSeg1_ = null;
      }
      break;
    case MODE_DEL:
      // This is silly. Ah well.
      mousePt_ = new Point();
      break;
    case MODE_EDIT:
      n = polygon1_.size();
      double dist = Double.POSITIVE_INFINITY;

      for ( i = 0; i < n; i++ ) {
	double thisDist;
	Point  thisPoint;
	thisPoint = polygon1_.getPoint(i);
	thisDist = ( (thisPoint.x() - x)*(thisPoint.x() - x) +
		     (thisPoint.y() - y)*(thisPoint.y() - y) );
	if (thisDist < dist) {
	  dist = thisDist;
	  if ( thisDist < 0.0003 ) {
	    savePt_ = new Point(thisPoint.x(),thisPoint.y());
	    mousePt_ = thisPoint;
	    editPoly_ = polygon1_;

	  }
	}
      }
      n = polygon2_.size();
      for ( i = 0; i < n; i++ ) {
	double thisDist;
	Point  thisPoint;
	thisPoint = polygon2_.getPoint(i);
	thisDist = ( (thisPoint.x() - x)*(thisPoint.x() - x) +
		     (thisPoint.y() - y)*(thisPoint.y() - y) );
	if (thisDist < dist) {
	  dist = thisDist;
	  if ( thisDist < 0.0003 ) {
	    savePt_ = new Point(thisPoint.x(),thisPoint.y());
	    mousePt_ = thisPoint;
	    editPoly_ = polygon2_;
	  }
	}
      }

    }
  }

  public void mouseDrag(double x, double y) {
    
    /**
      * If mousePt_ is null, means that move was started in
      * an illegal area.. So..
      */
    if (mousePt_ == null)
      return;

    switch (currentMode_) {
    case MODE_ADD1:
      if (polygon1_.checkNewConvex(x,y)) {
	mousePt_.x( x );
	mousePt_.y( y );
      }
      break;
    case MODE_ADD2:
      if (polygon2_.checkNewConvex(x,y)) {
	mousePt_.x( x );
	mousePt_.y( y );
      }
      break;
    case MODE_EDIT:
      double bakx = mousePt_.x();
      double baky = mousePt_.y();
      mousePt_.x( x );
      mousePt_.y( y );
      if (editPoly_ == null || !editPoly_.checkConvex(mousePt_)) {
	mousePt_.x( bakx );
	mousePt_.y( baky );
      }
      break;
    }
  }

  public void mouseUp(double x, double y) {

    int i;
    int n;

    /**
      * If mousePt_ is null, means that move was started in
      * an illegal area.. So..
      */
    if (mousePt_ == null)
      return;

    switch (currentMode_) {
    case MODE_ADD1:
      if (polygon1_.checkNewConvex(x,y)) {
	mousePt_.x( x );
	mousePt_.y( y );
      }
      try {
	polygon1_.addPoint(mousePt_);
      } catch (Exception e) {
	e.printStackTrace();
      }
      mousePt_ = null;
      curSeg1_ = curSeg2_ = null;

      break;
      
    case MODE_ADD2:
      if (polygon2_.checkNewConvex(x,y)) {
	mousePt_.x( x );
	mousePt_.y( y );
      }
      try {
	polygon2_.addPoint(mousePt_);
      } catch (Exception e) {
	e.printStackTrace();
      }
      mousePt_ = null;
      curSeg1_ = curSeg2_ = null;
      break;
      
    case MODE_EDIT:
      double bakx = mousePt_.x();
      double baky = mousePt_.y();
      mousePt_.x( x );
      mousePt_.y( y );
      if (editPoly_ == null || !editPoly_.checkConvex(mousePt_)) {
	mousePt_.x( bakx );
	mousePt_.y( baky );
      }
      mousePt_ = null;
      break;

    case MODE_DEL:
      mousePt_ = null;
      double dist = Double.POSITIVE_INFINITY;
      n = polygon1_.size();
      Point delPoint = null;
      Polygon delPolygon = null;

      for ( i = 0; i < n; i++ ) {
	double thisDist;
	Point  thisPoint;
	thisPoint = polygon1_.getPoint(i);
	thisDist = ( (thisPoint.x() - x)*(thisPoint.x() - x) +
		     (thisPoint.y() - y)*(thisPoint.y() - y) );
	if (thisDist < dist) {
	    dist = thisDist;
	  if ( thisDist < 0.0003 ) {
	    delPoint = thisPoint;
	    delPolygon = polygon1_;
	  }
	}
      }

      n = polygon2_.size();
      for ( i = 0; i < n; i++ ) {
	double thisDist;
	Point  thisPoint;
	thisPoint = polygon2_.getPoint(i);
	thisDist = ( (thisPoint.x() - x)*(thisPoint.x() - x) +
		     (thisPoint.y() - y)*(thisPoint.y() - y) );
	if (thisDist < dist) {
	    dist = thisDist;
	  if ( thisDist < 0.0003 ) {
	    delPoint = thisPoint;
	    delPolygon = polygon2_;
	  }
	}
      }
      if (delPolygon != null) {
	delPolygon.delPoint(delPoint);
	rebuildPolygon(delPolygon);
      }

      break;
    }
  }

  public void mouseCancel() {
    // Cancel recently added segments.
    switch (currentMode_) {
    case MODE_ADD1:
      if ( curSeg1_ != null ) {
	curSeg1_.getP1().remSeg(curSeg1_);
	curSeg1_= curSeg2_ = null;
	mousePt_ = null;
	if (polygon1_.size() > 1) {
	  Point last = polygon1_.getPoint(polygon1_.size()-1);
	  curSeg1_ = new Segment(last, polygon1_.getPoint(0), 1);
	  last.addSeg(curSeg1_);
	  curSeg1_ = null;
	}
      }
      break;
    case MODE_ADD2:
      if ( curSeg1_ != null ) {
	curSeg1_.getP1().remSeg(curSeg1_);
	curSeg1_= curSeg2_ = null;
	mousePt_ = null;
	if (polygon2_.size() > 1) {
	  Point last = polygon2_.getPoint(polygon2_.size()-1);
	  curSeg1_ = new Segment(last, polygon2_.getPoint(0), 2);
	  last.addSeg(curSeg1_);
	  curSeg1_ = null;
	}
      }
      break;

    case MODE_EDIT:
      if ( savePt_ != null && mousePt_ != null) {
	mousePt_.x(savePt_.x());
	mousePt_.y(savePt_.y());
	mousePt_ = null;
      }
      break;
    }
  }

  /**
    * Return the total number of points, not accounting for the fact
    * that most points on the intersection polygon will be repeated
    * in the two original polygons.
    */
  public int getNbPoints() {
    if ((currentMode_ == MODE_ADD1 || currentMode_ == MODE_ADD2) &&
	mousePt_ != null)
      return polygon1_.size() + polygon2_.size() + polygoni_.size() + 1;


    return polygon1_.size() + polygon2_.size() + polygoni_.size();
  }

  public Point getPoint(int i) {
    if ( i < polygon1_.size() )
      return polygon1_.getPoint(i);

     i -= polygon1_.size();
    if ( i < polygon2_.size() )
      return polygon2_.getPoint(i);

    i -= polygon2_.size();
    if ( i < polygoni_.size() )
      return polygoni_.getPoint(i);

    i -= polygoni_.size();
    if ( i == 0 && mousePt_ != null)
      return mousePt_;

    return null;
  }

  protected void rebuildPolygon(Polygon p) {

    int sz = p.size();
    int i;
    int ident = p.getIdent();

    if ( sz > 0) {
      //ident = p.getPoint(0).getIdent();
      p.getPoint(0).clrSegs();
      p.getPoint(0).setIdent(ident);
    }

    for (i=1; i<sz; i++) {
      Point pt = p.getPoint(i);
      pt.clrSegs();
      pt.setIdent(ident);
      Point last = p.getPoint(i-1);
      last.addSeg(new Segment(last, pt, ident));
    }

    if (sz > 2) {
      Point last = p.getPoint(sz-1);
      last.addSeg(new Segment(last, p.getPoint(0), ident));
    }
  }

  private final void clearIPolygon() {
    if ( polygoni_ == null ||
	 polygoni_.size() > 0 ) {
      polygoni_ = new Polygon();
      polygoni_.setIdent(6);
    }
    
  }

  private final void clearCvxHull() {
    /**
    if ( pocketLids_ != null ) {

      int i, n = pocketLids_.size();

      for (i=0; i < n; i++) {
	Segment s = (Segment)pocketLids_.elementAt(i);
	s.getP1().remSeg(s.getP2(),3);
      }
      pocketLids_ = null;
    }
    **/

    int i, n;
    n = polygon1_.size();
    for (i=0; i<n; i++) {
      int j;
      Point p = polygon1_.getPoint(i);      
      int m=p.nbSegs();
      for (j=1; j < m; j++)
	p.remSeg(1);
    }
    n = polygon2_.size();
    for (i=0; i<n; i++) {
      int j;
      Point p = polygon2_.getPoint(i);      
      int m=p.nbSegs();
      for (j=1; j < m; j++)
	p.remSeg(1);
    }

  }


  /**
    *  Returns false if the polygons don't have enough points (min 3 each)
    */
  public boolean computeCHinit() {
    

    if ( triangDone_ ) {
      rebuildPolygon(polygon1_);
      rebuildPolygon(polygon2_);
    }
    clearIPolygon();
    triangDone_ = false;
    xSectionBuilt_ = false;

    clearCvxHull();

    curPt1_ = curPt2_ = null;

    int sz1 = polygon1_.size();
    int sz2 = polygon2_.size();

    if (sz1 < 3 || sz2 < 3 )
      return false;

    curPt1idx_ = 0;
    curPt2idx_ = 0;
    gotAroundOnce1_ = false;
    gotAroundOnce2_ = false;
    pocketLids_ = new Vector();
    poly1LidIdx_= new int[sz1*2+1];
    poly2LidIdx_= new int[sz2*2+1];
    poly1XIdx_= new int[sz1*2+1];
    poly2XIdx_= new int[sz2*2+1];


    int i, n;

    //Find pt with min x coord on poly 1

    n = polygon1_.size();
    for (i=1; i<n; i++) {
      if (polygon1_.getPoint(i).x() < polygon1_.getPoint(curPt1idx_).x() )
	curPt1idx_ = i;
    }
    init1idx_ = curPt1idx_;


    //Find pt with min x coord on poly 2

    n = polygon2_.size();
    for (i=1; i<n; i++) {
      if (polygon2_.getPoint(i).x() < polygon2_.getPoint(curPt2idx_).x() )
	curPt2idx_ = i;
    }
    init2idx_ = curPt2idx_;

    Point v1 = new Point(polygon1_.getPoint((curPt1idx_+1)%sz1).x(),
			 polygon1_.getPoint((curPt1idx_+1)%sz1).y() );
    v1.minus(polygon1_.getPoint(curPt1idx_));
    v1.normalize();
    
    ang1_ = Math.acos(-v1.y());
    
    Point v2 = new Point(polygon2_.getPoint((curPt2idx_+1)%sz2).x(),
			 polygon2_.getPoint((curPt2idx_+1)%sz2).y() );
    v2.minus(polygon2_.getPoint(curPt2idx_));
    v2.normalize();
    
    ang2_ = Math.acos(-v2.y());


    ang1P_ = Math.PI/2;
    ang2P_ = Math.PI/2;


    Point prevpt1 = new Point(polygon1_.getPoint(curPt1idx_).x(),-2.);

    whichSideof1is2_ = Side.whichSide(polygon1_.getPoint(curPt1idx_),
				      prevpt1,
				      polygon2_.getPoint(curPt2idx_));


    curPt1_ = polygon1_.getPoint(curPt1idx_);
    curPt2_ = polygon2_.getPoint(curPt2idx_);

    return true;
  }


  /**
    * Returns true when CH is found.
    * Check if we're back past 0, we're in trouble..
    * Rather than a boolean, we should return an int that indicates
    * status, and corresponds to static constants.
    * @return XEngine.DONE    if the convex hull is found
    *         XEngine.OK      if the step has been done without errors
    *         XEngine.LOOPERR if it has been detected that one poly is
    *                         inside the other.
    */
  public int computeCHstep() {

    int sz1 = polygon1_.size();
    int sz2 = polygon2_.size();


    boolean movecal1 = false;
    boolean movecal2 = false;
    boolean done     = false;

    Point begPt1 = polygon1_.getPoint(curPt1idx_);
    Point begPt2 = polygon2_.getPoint(curPt2idx_);
    int begPt1idx = curPt1idx_;
    int begPt2idx = curPt2idx_;


    if ( ang1_ == ang2_ ) {
      movecal1 = true;
      movecal2 = true;
    }
    else if (ang1_ < ang2_) {
      ang2_ -= ang1_;
      ang2P_ -= ang1_;
      movecal1 = true;
    }
    else {			// ang2_ < ang1_
      ang1_ -= ang2_;
      ang1P_ -= ang2_;
      movecal2 = true;
    }


    if ( movecal1 ) {

      ang1P_ -= ang1_;

      Point prevpt1 = polygon1_.getPoint(curPt1idx_);
      curPt1idx_ = (curPt1idx_ + 1) % sz1;
      if (curPt1idx_ == (init1idx_ +1)%sz1 ) {
	if ( gotAroundOnce1_ && gotAroundOnce2_ ) {
	  if ( pocketLids_.size() > 0 )
	    done = true;
	  else {
	    curPt1_ = curPt2_ = null;
	    return LOOPERR;
	  }
	}
	gotAroundOnce1_ = true;
      }
      
      Point v1 = new Point(polygon1_.getPoint((curPt1idx_+1)%sz1).x(),
			   polygon1_.getPoint((curPt1idx_+1)%sz1).y() );
      v1.minus(polygon1_.getPoint(curPt1idx_));
      v1.normalize();

      Point v2 = new Point(polygon1_.getPoint(curPt1idx_).x(),
			   polygon1_.getPoint(curPt1idx_).y() );
      v2.minus(prevpt1);
      v2.normalize();

      ang1_ = Math.acos(v1.x()*v2.x() + v1.y()*v2.y());
 
    }
    
    if ( movecal2 ) {

      ang2P_ -= ang2_;

      Point prevpt2 = polygon2_.getPoint(curPt2idx_);
      curPt2idx_ = (curPt2idx_ + 1) % sz2;
      if (curPt2idx_ == (init2idx_ + 1)%sz2 ) {
	if ( gotAroundOnce1_ && gotAroundOnce2_ ) {
	  if ( pocketLids_.size() > 0 )
	    done = true;
	  else {
	    curPt1_ = curPt2_ = null;
	    return LOOPERR;
	  }
	}
	gotAroundOnce2_ = true;
      }


      Point v1 = new Point(polygon2_.getPoint((curPt2idx_+1)%sz2).x(),
			   polygon2_.getPoint((curPt2idx_+1)%sz2).y() );
      v1.minus(polygon2_.getPoint(curPt2idx_));
      v1.normalize();

      Point v2 = new Point(polygon2_.getPoint(curPt2idx_).x(),
			   polygon2_.getPoint(curPt2idx_).y() );
      v2.minus(prevpt2);
      v2.normalize();

      ang2_ = Math.acos(v1.x()*v2.x() + v1.y()*v2.y());

    }

    Point p1b = new Point( Math.cos(ang1P_), - Math.sin(ang2P_));

    p1b.add(polygon1_.getPoint(curPt1idx_));

    int newside = Side.whichSide(polygon1_.getPoint(curPt1idx_),
				 p1b,
				 polygon2_.getPoint(curPt2idx_));


    if (whichSideof1is2_ != newside) {

      whichSideof1is2_ = newside;

      Point p1 = polygon1_.getPoint(curPt1idx_);
      Point p2 = polygon2_.getPoint(curPt2idx_);


      Segment lid = new Segment(begPt1,begPt2, 3);
      begPt1.addSeg(lid);

      poly1LidIdx_[pocketLids_.size()] = begPt1idx;
      poly2LidIdx_[pocketLids_.size()] = begPt2idx;
      pocketLids_.addElement(lid);


    }

    if (done) {
      curPt1_ = curPt2_ = null;
      if ( ((Segment)pocketLids_.firstElement()).equals(
	    (Segment)pocketLids_.lastElement()) ) {
	pocketLids_.removeElementAt(pocketLids_.size()-1);
      }
          
      return DONE;
    }
    else {
      curPt1_ = polygon1_.getPoint(curPt1idx_);
      curPt2_ = polygon2_.getPoint(curPt2idx_);
      return OK;
    }
  }


  public void resetTriang() {
    clearIPolygon();
    triangReady_ = false;
    triangDone_ = false;
    xSectionBuilt_ = false;
  }

  public boolean initTriang() {
    if ( pocketLids_.size() > 0 ) {
      currentPocketNb_ = 0;
      triangReady_ = true;
      polygoni_ = new Polygon();
      polygoni_.setIdent(6);
      currentLid_ = (Segment)pocketLids_.elementAt(currentPocketNb_);

      curPt1idx_ = 0;
      curPt2idx_ = 0;

      int i, n;

      n = polygon1_.size();
      for (i=0; i< n; i++) {
	Point pt = polygon1_.getPoint(i);
	pt.fwIsXchain(false);
	pt.rvIsXchain(false);
      }
      n = polygon2_.size();
      for (i=0; i< n; i++) {
	Point pt = polygon2_.getPoint(i);
	pt.fwIsXchain(false);
	pt.rvIsXchain(false);
      }

      return true;
    }
    else
      return false;
  }

  public Point getCurPocketPos() {
    if (!triangReady_)
      return null;
    /**
       Point p1 = currentLid_.getP1();
       Point p2 = currentLid_.getP2();
       
       return new Point(p1.x() + (p2.x() - p1.x())/2,
       p1.y() + (p2.y() - p1.y())/2 );
    **/
    int i1 = poly1XIdx_[currentPocketNb_-1];
    int i2 = poly2XIdx_[currentPocketNb_-1];

    Point p1 = polygon1_.getPoint(i1);
    Point p2 = polygon2_.getPoint(i2);

    Point n1 = null, n2 = null;
    if (p1Iscwise_) {
      n1 = polygon1_.getPoint((i1+1)%polygon1_.size());
      n2 = polygon2_.getPoint((i2+1)%polygon2_.size());
    }
    else {
      n2 = polygon2_.getPoint((i2+1)%polygon2_.size());
      n1 = polygon1_.getPoint((i1+1)%polygon1_.size());
    }

   Point min = new Point( Math.min( n1.x(),
				     Math.min( n2.x(),
					       Math.min( p1.x(), p2.x()))),
			   Math.min( n1.y(),
				     Math.min( n2.y(),
					       Math.min( p1.y(), p2.y()))));
   Point max = new Point( Math.max( n1.x(),
				     Math.max( n2.x(),
					       Math.max( p1.x(), p2.x()))),
			   Math.max( n1.y(),
				     Math.max( n2.y(),
					       Math.max( p1.y(), p2.y()))));

    
   return new Point( max.x()/2. + min.x()/2., max.y()/2. + min.y()/2. ); 


  }

  /**
    * Returns square bbox of last xsection.
    */
  public double getCurPocketSz() {
    if ( !triangReady_ )
      return 1.;

    int i1 = poly1XIdx_[currentPocketNb_-1];
    int i2 = poly2XIdx_[currentPocketNb_-1];

    Point p1 = polygon1_.getPoint(i1);
    Point p2 = polygon2_.getPoint(i2);

    Point n1 = null, n2 = null;
    if (p1Iscwise_) {
      n1 = polygon1_.getPoint((i1+1)%polygon1_.size());
      n2 = polygon2_.getPoint(decrIdx(i2,polygon2_.size()));
    }
    else {
      n2 = polygon2_.getPoint((i2+1)%polygon2_.size());
      n1 = polygon1_.getPoint(decrIdx(i1,polygon1_.size()));
    }

    Point min = new Point( Math.min( n1.x(),
				     Math.min( n2.x(),
					       Math.min( p1.x(), p2.x()))),
			   Math.min( n1.y(),
				     Math.min( n2.y(),
					       Math.min( p1.y(), p2.y()))));
    Point max = new Point( Math.max( n1.x(),
				     Math.max( n2.x(),
					       Math.max( p1.x(), p2.x()))),
			   Math.max( n1.y(),
				     Math.max( n2.y(),
					       Math.max( p1.y(), p2.y()))));
    
    return Math.max( max.x() - min.x(), max.y() - min.y());
  }



  /**
    * Returns TRUE if a next intersection was computed,
    * or false if there are no more intersections (or there are
    * no intersections at all).
    */
  public int computeNextXSection() {

    //We found the X when the point on poly2 ain't on the
    //same side as the lid anymore.

    Polygon ourPoly   = null;
    Polygon otherPoly = null;
    int ourIdx   = 0;
    int otherIdx = 0;
    int ourSz    = 0;
    int otherSz  = 0;

    if (currentPocketNb_ >= pocketLids_.size() ) {
      triangDone_ = true;
      return DONE;
    }

    Segment lid = (Segment)pocketLids_.elementAt(currentPocketNb_);

    //Determine if lid starts with poly 1 or poly 2
    if (Side.whichSide(lid.getP2(), lid.getP1(),
		       polygon1_.getPoint(
			 (poly1LidIdx_[currentPocketNb_]+1)%polygon1_.size()) )
	== Side.UNDER ) {
      //It starts with poly1
      if (currentPocketNb_ == 0)
	p1IsCwiseOn1stLid_ = true;
      ourPoly   = polygon1_;
      otherPoly = polygon2_;
      ourSz     = polygon1_.size();
      otherSz   = polygon2_.size();
      ourIdx    = poly1LidIdx_[currentPocketNb_];
      otherIdx  = poly2LidIdx_[currentPocketNb_];
      
    }
    else {
      //It starts with poly2
      if (currentPocketNb_ == 0)
	p1IsCwiseOn1stLid_ = false;
      ourPoly   = polygon2_;
      otherPoly = polygon1_;
      ourSz     = polygon2_.size();
      otherSz   = polygon1_.size();
      ourIdx    = poly2LidIdx_[currentPocketNb_];
      otherIdx  = poly1LidIdx_[currentPocketNb_];
    }

    /**
      * Toussaint's alg is here texto, except for a small part where
      * we detect if we've been looping or not.
      */

    boolean done = false;
    int ourStartIdx   = ourIdx;
    int otherStartIdx = otherIdx;

    while (!done) {

      done = true;
      
      while ( Side.whichSide( ourPoly.getPoint(ourIdx),
			      ourPoly.getPoint((ourIdx+1)%ourSz),
			      otherPoly.getPoint(decrIdx(otherIdx,otherSz))
			      ) == Side.UNDER ) {
	otherIdx = decrIdx(otherIdx,otherSz);

	if ( otherIdx == otherStartIdx ) {
	  curPt1_ = curPt2_ = null;
	  return NOXERR;
	}
	done = false;
      }
      
      while ( Side.whichSide( otherPoly.getPoint(otherIdx),
			      otherPoly.getPoint(decrIdx(otherIdx,otherSz)),
			      ourPoly.getPoint((ourIdx+1)%ourSz)
			      ) == Side.OVER ) {
	ourIdx = (ourIdx+1)%ourSz;

	if ( ourIdx == ourStartIdx ) {
	  curPt1_ = curPt2_ = null;
	  return NOXERR;
	}

	done = false;
      }
    }

    //ourPoly.getPoint(ourIdx).getSeg(0).setIdent(4);
    ourPoly.getPoint(ourIdx).fwIsXchain(true);

    //otherPoly.getPoint(decrIdx(otherIdx,otherSz)).getSeg(0).setIdent(4);
    otherPoly.getPoint(decrIdx(otherIdx,otherSz)).rvIsXchain(true);


    

    if (ourPoly == polygon1_) {
      poly1XIdx_[currentPocketNb_] = ourIdx;
      poly2XIdx_[currentPocketNb_] = decrIdx(otherIdx,otherSz);
      p1Iscwise_ = true;
    }
    else {
      poly2XIdx_[currentPocketNb_] = ourIdx;
      poly1XIdx_[currentPocketNb_] = decrIdx(otherIdx,otherSz);
      p1Iscwise_ = false;
    }

    currentLid_ = lid;  
    currentPocketNb_++;

    return OK;
  }

  public void showLastXSection() {
    //ourPoly.getPoint(ourIdx).getSeg(0).setIdent(4);
    //otherPoly.getPoint(decrIdx(otherIdx,otherSz)).getSeg(0).setIdent(4);
    polygon1_.getPoint(poly1XIdx_[currentPocketNb_-1]).getSeg(0).setIdent(4);
    polygon2_.getPoint(poly2XIdx_[currentPocketNb_-1]).getSeg(0).setIdent(4);
  }

  public boolean buildXsect() {
    if (!triangDone_)
      return false;

    int i, n;
    
    /**
      * Find which in chain is in clockwise dir (poly 1 or 2)
      * for first lid.
      *    Starting from a lid pt on poly1. If what we hit first is
      *    a rvIsXchain pt, poly1 is clockwise. If what we hit
      *    first is another 2 segment pt, then poly1 is the c-clockwise.
      */

    Polygon clockwise = null;
    Polygon cClockwise = null;
    int[] cwIdx = null;
    int[] cCwIdx = null;

    boolean nowWeKnow = false;
    i=poly1LidIdx_[0];

    if ( p1IsCwiseOn1stLid_ ) {
      clockwise = polygon1_;
      cClockwise = polygon2_;
      cwIdx = poly1LidIdx_;
      cCwIdx = poly2LidIdx_;
    }
    else {
      clockwise = polygon2_;
      cClockwise = polygon1_;
      cwIdx = poly2LidIdx_;
      cCwIdx = poly1LidIdx_;
    }

    for ( i=0, n=pocketLids_.size(); i < n; i++) {
      
      /**
	* Get the two points (fwIsXchain on clockwise poly and
	* rvIsXchain on c-clockwise) by searching from the lid pts.
	*/

      int j1=0, j2=0;
      Point p1 = null;
      Point p2 = null;
      if (polygon1_ == clockwise) {
	j1 = poly1XIdx_[i];
	p1 = polygon1_.getPoint(j1);
	j2 = poly2XIdx_[i];
	p2 = polygon2_.getPoint(j2);
      }
      else {
	j1 = poly2XIdx_[i];
	p1 = polygon2_.getPoint(j1);
	j2 = poly1XIdx_[i];
	p2 = polygon1_.getPoint(j2);
      }
      
      /**
	* Find xsection on those two segments (take seg0).
	*/
      Point p = clockwise.getPoint((j1+1)%clockwise.size());
      Point t1 = new Point(p.x(), p.y());
      t1.minus(p1);
      //t1.normalize();
      p = cClockwise.getPoint((j2+1)%cClockwise.size());
      Point t2 = new Point(p.x(), p.y());
      t2.minus(p2);
      //t2.normalize();
      Point pi = lineXsection(p1,t1,p2,t2);
      pi.setIdent(6);
      polygoni_.addPointNoCheck(pi);

      
      /**
	* Add points on that chain until we find a point that
	* is a rvIsXChain (and add that last point)
	*/
      
      while (!p1.rvIsXchain()) {
	  j1 = (j1+1)%clockwise.size();
	  p1 = clockwise.getPoint(j1);
	  pi =  new Point(p1.x(),p1.y());
	  pi.setIdent(6);
	  polygoni_.addPointNoCheck(pi);
      }

      /**
	* Switch the clockwise and c-clockwise polys.
	*/
      Polygon pSwap;
      int[] iSwap;

      pSwap = clockwise;
      clockwise = cClockwise;
      cClockwise = pSwap;

      iSwap = cwIdx;
      cwIdx = cCwIdx;
      cCwIdx = iSwap;
    }

    rebuildPolygon(polygoni_);

    xSectionBuilt_ = true;
    clearCvxHull();
    return true;
  }


  public boolean xSectionBuilt() {
    return xSectionBuilt_;
  }


  /** Decrement an index, with the special case of
    * decr(0,n) == n-1. (Just like modulo addition).
    */
  private final int decrIdx(int i, int n) {
    if (i != 0)
      return i-1;
    else
      return n-1;
  }

  /**
    * Intersects two lines in parametric form.
    */
  private final Point lineXsection( Point p1, Point t1,
				    Point p2, Point t2 ) {

    double a = p1.x();
    double b = p1.y();
    double c = t1.x();
    double d = t1.y();
    double e = p2.x();
    double f = p2.y();
    double g = t2.x();
    double h = t2.y();

    double y;
    
    y = - (b*c-a*d+d*e-c*f)/(d*g-c*h);
    

    if ( y == Double.POSITIVE_INFINITY ||
	 y == Double.NEGATIVE_INFINITY   ) {
      throw new RuntimeException("Huh? No intersection! Parallel lines??");
    }

    Point p = new Point( t2.x(), t2.y() );
    p.scale(y);
    p.add(p2);
    return p;

  }

}
