package ep.cg.widgets;

import java.awt.*;
import java.util.Vector;
import ep.cg.ModeChangeListener;
import ep.cg.ActionListener;

/**
  * A widget which uses preset images to display a choice
  * of actions,and which allows to select an action
  * by clicking on it.
  * @author ric Plante
  **/

public class ActionPanel extends Canvas {

  private Vector listeners_;

  private Image nrm_;
  private Image int_;
  private Image clr_;
  private Image qit_;

  private Image current_;

  private boolean clickedOn_ = false;

  public ActionPanel(Image nrm, Image intr, Image clr, Image qit) {
    listeners_ = new Vector();
    nrm_ = current_ = nrm;
    int_ = intr;
    clr_ = clr;
    qit_ = qit;
    resize(320,151);
  }

  public void paint(Graphics g) {
    update(g);
  }

  public void update(Graphics g) {

    g.drawImage( current_, 0, 0, this );
  }

  public Dimension minimumSize() {
    return new Dimension(320,151);
  }

  public Dimension preferredSize() {
    return minimumSize();
  }

  public boolean mouseDown(Event evt, int x, int y) {
    clickedOn_ = true;
    setModeAndImage(x,y);
    return true;
  }

  public boolean mouseDrag(Event evt, int x, int y) {
    if (!clickedOn_)
      //return super.handleEvent(evt);
      return false;
    setModeAndImage(x,y);
    return true;
  }

  public boolean mouseUp(Event evt, int x, int y) {
    if (clickedOn_) {
      /* This is an action that started in our canvas, and it
       * just ended. We reset our flag.
       */
      clickedOn_ = false;
      if (current_ == int_ )
	notifyListeners(ActionListener.ACTION_INT);
      else if (current_ == clr_ )
	notifyListeners(ActionListener.ACTION_CLR);
      else if (current_ == qit_ )
	notifyListeners(ActionListener.ACTION_QIT);
      current_ = nrm_;
      repaint(0);
      return true;
    }
    else
      //return super.handleEvent(evt);
      return false;
  }

  public boolean mouseExit(Event evt, int x, int y) {
    if (clickedOn_) {
      clickedOn_ = false;
      current_  = nrm_;
      repaint(0);
      return true;
    }
    else
      //return super.handleEvent(evt);
      return false;
  }

  private void setModeAndImage(int x, int y) {    
    if (x > 162) {
      if (y > 78)
	current_ = qit_;
    }
    else {
      if (y > 78)
	current_ = clr_;
      else
	current_ = int_;
    }
    repaint(0);
  }

  public void addListener(ActionListener l) {
    if ( listeners_.indexOf(l) == -1 )
      listeners_.addElement(l);
  }

  public void removeListener(ActionListener l) {
    listeners_.removeElement(l);
  }

  public void notifyListeners(int act) {
    int size = listeners_.size();
    int i;

    for (i=0; i<size; i++) {
      ((ActionListener)listeners_.elementAt(i)).actionEvent(act);
    }
  }
}
