package ep.cg.widgets;

import java.awt.*;
import java.util.Vector;
import ep.cg.ModeChangeListener;

/**
  * A widget which uses preset images to display a choice
  * of edit modes, the select mode, and which allows to change
  * the mode by clicking on it.
  * @author ric Plante
  **/

public class ModePanel extends Canvas {

  private Vector listeners_;

  private Dimension offDimension_;
  private Image offImage_;

  private Image add1_;
  private Image add2_;
  private Image edit_;
  private Image del_; 
  private Image current_;
  private Image last_;
  private Image lastActive_;

  private boolean clickedOn_ = false;

  public ModePanel(Image add1, Image add2, Image edit, Image del) {
    listeners_ = new Vector();
    add1_ = current_ = add1;
    add2_ = add2;
    edit_ = edit;
    del_  = del;
    resize(320,151);
  }

  public void paint(Graphics g) {
    update(g);
  }

  public void update(Graphics g) {
    /*
    Dimension d = size();

    if ( (offImage_ == null) ||
	 (size.width != offDimension_.width) ||
	 (size.height != offDimension_.height) ) {
      offDimension_ = d;
      offImage_ = createImage(d.width, d.height);
    }
    */
    g.drawImage( current_, 0, 0, this );
  }

  public Dimension minimumSize() {
    return new Dimension(320,151);
  }

  public Dimension preferredSize() {
    return minimumSize();
  }

  public boolean mouseDown(Event evt, int x, int y) {
    clickedOn_ = true;
    lastActive_ = current_;
    last_ = current_;
    setModeAndImage(x,y);
    return true;
  }

  public boolean mouseDrag(Event evt, int x, int y) {
    if (!clickedOn_)
      //return super.handleEvent(evt);
      return false;
    setModeAndImage(x,y);
    return true;
  }

  public boolean mouseUp(Event evt, int x, int y) {
    if (clickedOn_) {
      /* This is an action that started in our canvas, and it
       * just ended. We reset our flag.
       */
      clickedOn_ = false;
      if (current_ == add1_ )
	notifyListeners(ModeChangeListener.MODE_ADD1);
      else if (current_ == add2_ )
	notifyListeners(ModeChangeListener.MODE_ADD2);
      else if (current_ == edit_ )
	notifyListeners(ModeChangeListener.MODE_EDIT);
      else if (current_ == del_ )
	notifyListeners(ModeChangeListener.MODE_DEL);
      return true;
    }
    else
      //return super.handleEvent(evt);
      return false;
  }

  public boolean mouseExit(Event evt, int x, int y) {
    if (clickedOn_) {
      clickedOn_ = false;
      current_  = lastActive_;
      repaint(0);
      return true;
    }
    else
      //return super.handleEvent(evt);
      return false;
  }

  private void setModeAndImage(int x, int y) {    
    Image lastCurrent_ = current_;
    if (x > 162) {
      if (y > 78)
	current_ = del_;
      else
	current_ = edit_;
    }
    else {
      if (y > 78)
	current_ = add2_;
      else
	current_ = add1_;
    }
    if ( last_ != current_ )
      repaint(0);
    last_ = lastCurrent_;
    
  }

  public void addListener(ModeChangeListener l) {
    if ( listeners_.indexOf(l) < 0 )
      listeners_.addElement(l);
  }

  public void removeListener(ModeChangeListener l) {
    listeners_.removeElement(l);
  }

  public void notifyListeners(int mode) {
    int size = listeners_.size();
    int i;

    for (i=0; i<size; i++) {
      ((ModeChangeListener)listeners_.elementAt(i)).modeChangeEvent(mode);
    }
  }
}
