package ep.cg.widgets;

import java.awt.Graphics;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Event;
import java.awt.Font;
import java.util.Vector;
import ep.cg.*;

/**
  * The canvas on which line segments are drawn. Contains an instance
  * of XEngine, which is used to determine what is drawable and what is
  * drawn.
  * @author ric Plante
  **/

public class PolyCanvas extends Canvas
implements ModeChangeListener, ActionListener, Runnable {

  private Image dot1_;
  private Image dot2_;
  private Image doti_;

  private double zoom_ = 1;
  private double toZoom_ = 1;
  private double xoff_ = 0;
  private double yoff_ = 0;
  private static final int ZOOM_STEPS_ = 16;

  private int currentMode_ = MODE_ADD1;
  private boolean clickedOn_ = false;
  private static Color SEP_CLR = new Color((float)0.8,(float)0.7,(float)0.3);

  private static XEngine ENGINE = new XEngine();


  private boolean xsecting_ = false;
  private boolean xsected_ = false;

  private String currentMsg_ = null;

  public PolyCanvas(Image dot1, Image dot2, Image doti) {
    super();
    setBackground(Color.black);
    dot1_ = dot1;
    dot2_ = dot2;
    doti_ = doti;
    currentMsg_ = "Welcome to the Linear Convex Polygon Intersector.";
  }

  public PolyCanvas() {
    super();
    dot1_ = dot2_ = doti_ = null;
  }


  protected void reset() {
    ENGINE.reset();
    clickedOn_ = false;
    xsecting_ = false;
    xsected_ = false;
  }

  public synchronized void run() {

    xsected_ = false;
    xsecting_ = true;
    if (!ENGINE.computeCHinit()) {
      currentMsg_ = "Each polygon should have at least 3 points!";
      xsecting_ = false;
      return;
    }
    
    int ret;
    currentMsg_ = "Computing convex hull with rotating calipers...";
    while ( (ret = ENGINE.computeCHstep()) == ENGINE.OK ) {
      repaint(0);
      Thread.yield();
      try {
	Thread.sleep(100);
      } catch (InterruptedException ie);
    }
    
    xsecting_ = false;
    
    if (ret == ENGINE.LOOPERR) {
      currentMsg_ = "The polygons are one inside another!";
      xsected_ = false;
    }
    else {
      
      repaint(0);
      Thread.yield();
      try {
	Thread.sleep(100);
      } catch (InterruptedException ie);
      
      
      if ( !ENGINE.initTriang() )
	xsected_ = false;
      else {
	int ret2;
	currentMsg_ = "Finding intersections (bottom of pockets)...";
	while ( (ret2 = ENGINE.computeNextXSection()) == ENGINE.OK ) {
	  
	  double ps = ENGINE.getCurPocketSz();
	  Point  pp = ENGINE.getCurPocketPos();
	  
	  double dx = (pp.x() - ps/2) / ZOOM_STEPS_;
	  double dy = (pp.y() - ps/2) / ZOOM_STEPS_;
	  
	  double ss = ( 1./ps) / ZOOM_STEPS_;
	  
	  int k,l;
	  for (l=0; l<2; l++) {
	    for ( k = 1; k < ZOOM_STEPS_; k++ ) {
	      zoom_ += ss;
	      /**
		 xoff_ += dx;
		 yoff_ += dy;
	      **/
	      xoff_ = pp.x() - ps/2;
	      yoff_ = pp.y() - ps/2;
	      
	      if (xoff_ < 0) xoff_ = 0;
	      else
		if ((xoff_)+ 1./zoom_ > 1.)
		  xoff_ = Math.max(1. - 1./zoom_, 0.);
	      
	      if (yoff_ < 0) yoff_ = 0;
	      else
		if ((yoff_)+ 1./zoom_ > 1.)
		  yoff_ = Math.max(1. - 1./zoom_, 0.);
	      
	      repaint(0);
	      Thread.yield();
	      try {
		Thread.sleep(34);
	      } catch (InterruptedException ie);
	    }
	    
	    if (l==0)
	      ENGINE.showLastXSection();
	    repaint(0);
	    Thread.yield();
	    try {
	      Thread.sleep(500);
	    } catch (InterruptedException ie);
	    ss = -ss;
	    dx = -dx;
	    dy = -dy;
	    //pp.x(0.5); pp.y(0.5);
	  }
	  xoff_ = yoff_ = 0;
	  
	  try {
	    Thread.sleep(800);
	  } catch (InterruptedException ie);
	}
	
	xsected_ = true;
	
	if ( ret2 == ENGINE.NOXERR ) {
	  currentMsg_ = "No intersection found!";
	  xsecting_ = false;
	}
	else {
	  currentMsg_ = "Building intersection polygon...";	
	  repaint(0);
	  Thread.yield();
	  if (!ENGINE.buildXsect())
	    //System.out.println("Hmm, trying to build without compute?");
	    // Go on... then.
	    xsecting_ = false;
	  currentMsg_ = "Done!";		  
	}
      }
    }

    xsecting_ = false;
    repaint(0);
    Thread.yield();
  }

  public void modeChangeEvent(int mode) {
    currentMsg_ = null;
    ENGINE.setMode(mode);
    if (!xsecting_ && xsected_) {
      xsected_ = false;
      repaint(0);
      Thread.yield();
    }
    currentMode_ = mode;
  }

  public void actionEvent(int act) {
    currentMsg_ = null;
    switch (act) {
    case ACTION_CLR:
      reset();
      repaint(0);
      Thread.yield();
      break;
    case ACTION_INT:
      if (!xsecting_ && xsected_) {
	xsected_ = false;
	repaint(0);
	Thread.yield();
      }
      Thread t = new Thread(this);
      t.start();
      break;
    }
  }


  public boolean mouseDown(Event evt, int x, int y) {
    
    currentMsg_ = null;
    if ( !xsecting_ )
      xsected_ = false;
    
    Dimension sz = size();

    double x_ = (((double)x) / ((double)sz.width) ) / zoom_ + xoff_;
    double y_ = (((double)y) / ((double)sz.height)) / zoom_ + yoff_;

    clickedOn_ = true;
    
    ENGINE.mouseDown( x_, y_ );
    repaint(0);
    Thread.yield();

    return true;
  }

  public boolean mouseDrag(Event evt, int x, int y) {
    if (!clickedOn_)
      return false;

    Dimension sz = size();
    double x_ = (((double)x) / ((double)sz.width) ) / zoom_ + xoff_;
    double y_ = (((double)y) / ((double)sz.height)) / zoom_ + yoff_;


    ENGINE.mouseDrag(x_, y_);
    repaint(0);
    Thread.yield();

    return true;
  }

  public boolean mouseUp(Event evt, int x, int y) {
    if (clickedOn_) {
      /* This is an action that started in our canvas, and it
       * just ended. We reset our flag.
       */
      Dimension sz = size();

      double x_ = (((double)x) / ((double)sz.width) ) / zoom_ + xoff_;
      double y_ = (((double)y) / ((double)sz.height)) / zoom_ + yoff_;
      clickedOn_ = false;

      ENGINE.mouseUp(x_, y_);
      repaint();
      Thread.yield();

      return true;
    }

    return false;
  }


  public boolean mouseExit(Event evt, int x, int y) {
    if (clickedOn_) {
      clickedOn_ = false;
      ENGINE.mouseCancel();
      repaint(0);
      Thread.yield();
      return true;
    }
    else
      return false;
  }
  
  public void paint(Graphics g) {
    update(g);
  }


  /**
    * TODO: Implement double-buffering.
    */
  Dimension odim_;
  Image oi_;
  Graphics og_;
  public void update(Graphics g) {

    int nbPoints_ = ENGINE.getNbPoints();
    Dimension dim = size();

    if ( (og_ == null)
	 || (dim.width != odim_.width)
	 || (dim.height != odim_.height) ) {
      odim_ = dim;
      oi_ = createImage(dim.width, dim.height);
      og_ = oi_.getGraphics();
    }

    og_.setColor(Color.black);
    og_.fillRect(0,0,dim.width,dim.height);

    og_.setFont(new Font("Helvetica",Font.BOLD ,12));

    int i;

    for (i=0; i < nbPoints_; i++) {
      Point p = ENGINE.getPoint(i);
      if (p == null)
	continue;

      int j;
      int n = p.nbSegs();

      int px = (int)(((p.x()-xoff_) * dim.width ) * zoom_ );
      int py = (int)(((p.y()-yoff_) * dim.height ) * zoom_ );

      for (j=0; j < n; j++) {
	Segment s = p.getSeg(j);
	if ( ENGINE.xSectionBuilt() )
	  if (s.getIdent() == 6)
	    og_.setColor(Color.green.darker());
	  else
	    og_.setColor(Color.darkGray);
	else
	  switch (s.getIdent()) {
	  case 1:
	    og_.setColor(Color.red);
	    break;
	  case 2:
	    og_.setColor(Color.yellow);
	    break;
	  case 3:
	    og_.setColor(Color.green);
	    break;
	  case 4:
	    og_.setColor(Color.blue);
	    break;
	  default:
	    og_.setColor(Color.magenta);
	  }
	og_.drawLine( px, py,
		    (int)(((s.getP2().x()-xoff_) * dim.width ) * zoom_),
		    (int)(((s.getP2().y()-yoff_) * dim.height) * zoom_) );
      }

      if ( !xsected_ ) {
	switch ( p.getIdent()) {
	case 1:
	  og_.drawImage(dot1_, px-8,py-8, this);
	  break;
	case 2:
	  og_.drawImage(dot2_, px-8,py-8, this);
	  break;
	case 3:
	  og_.drawImage(doti_, px-8,py-8, this);
	  break;
	}
      }
      /**
      if (ENGINE.xSectionBuilt()) {
	if ( p.getIdent() == 6) {
	  og_.drawImage(doti_, px-8,py-8, this);
	}
      }
      **/
    }
    
    
    int x, y;
    
    
    if ( ENGINE.curPt1_ != null) {
      og_.setColor(Color.red.darker().darker());
      x = (int)(((ENGINE.curPt1_.x()-xoff_) * dim.width ) * zoom_);
      y = (int)(((ENGINE.curPt1_.y()-yoff_) * dim.height)  * zoom_);
      
      og_.drawLine( x, y,
		  x+ (int)(1000*Math.cos(ENGINE.ang1P_)),
		  y- (int)(1000*Math.sin(ENGINE.ang1P_)) );
    }

    if ( ENGINE.curPt2_ != null) {
      og_.setColor(Color.yellow.darker().darker());
      x = (int)(((ENGINE.curPt2_.x()-xoff_) * dim.width ) * zoom_);
      y = (int)(((ENGINE.curPt2_.y()-yoff_) * dim.height) * zoom_);
      
      og_.drawLine( x, y,
		  x+ (int)(1000*Math.cos(ENGINE.ang2P_)),
		  y- (int)(1000*Math.sin(ENGINE.ang2P_)) );
    }

    if ( currentMsg_ != null ) {
      og_.setColor(Color.yellow);
      og_.drawString(currentMsg_, 5, dim.height - 5 );
    }

    og_.setColor(SEP_CLR);
    og_.fill3DRect(0,0,4,dim.height,true);
    g.drawImage( oi_, 0, 0, this);
  }
}
